unit BufferedFile;

interface

uses Classes, SysUtils;

type
  EFileError = Exception;

  TCharArray = array [0..0] of Char;
  PCharArray = ^TCharArray;

  TBufferStream = class(TStream)
  private
    { Buffer }
    FCurrentSize: LongInt;
    FBufferSize: LongInt;
    FBuffer: PCharArray;
		FBufferPos: LongInt;
    FReadSize: LongInt;
    FVirtualPos: LongInt;
    FHasWritten: Boolean;
    FFilled: Boolean;
    FHasReadSize: Boolean;
    procedure CheckSize;
    function GetCurrentPos: LongInt;
    function GetBOF: Boolean;
    function GetEOF: Boolean;
  protected
  	constructor Create(BufferSize: Integer);
    function DirectRead(var Buffer; Count: LongInt): LongInt; virtual; abstract;
    procedure DirectWrite(const Buffer; Count: LongInt); virtual; abstract;
    function GetSize: LongInt; virtual; abstract;
    procedure WriteBuffer;
    procedure FillBuffer;
    procedure CheckBuffer(Length: LongInt);
    procedure CheckEnd(Length: LongInt);
    {}
    procedure SetBufferSize(NewSize: Integer);
    procedure SetSize(NewSize: LongInt); override;
    {}
    property BufferPos: LongInt read FBufferPos;
    property VirtualPos: LongInt read FVirtualPos;
  public
		destructor Destroy; override;
		function Seek(Offset: LongInt; Origin: Word): LongInt; override;
    function Read(var Buffer; Count: LongInt): LongInt; override;
    function Write(const Buffer; Count: LongInt): LongInt; override;
    {}
    procedure FlushBuffer;
    procedure InvalidateBuffer;
    {}
    function ReadChar: Char;
    procedure ReadLn(var S : string);
    procedure WriteChar(Char: Char);
    procedure WriteLn(const Str: string);
    {}
    property BOF: Boolean read GetBOF;
    property EOF: Boolean read GetEOF;
    property BufferSize: Integer read FBufferSize write SetBufferSize;
  end;

  TFileBufferStream = class(TBufferStream)
  private
    FHandle: Integer;
  protected
    function DirectRead(var Buffer; Count: LongInt): LongInt; override;
    procedure DirectWrite(const Buffer; Count: LongInt); override;
    function GetSize: LongInt; override;
  public
    constructor Create(const FileName: string; Mode: Word);
    destructor Destroy; override;
  end;

implementation

function IMin(I1, I2: Integer): Integer;
begin
  if I1 > I2 then Result:= I2 else Result:= I1;
end;

function IMax(I1, I2: Integer): Integer;
begin
  if I1 < I2 then Result:= I2 else Result:= I1;
end;

{ ***************************************************************************** }
{ ******************************* TBufferStream ******************************* }
{ ***************************************************************************** }

{ ******************** Cration, ouverture, fermeture ********************** }
constructor TBufferStream.Create(BufferSize: Integer);
begin
	inherited Create;
  FBufferSize:= BufferSize;
  GetMem(FBuffer, FBufferSize);
  FCurrentSize:= GetSize;
end;

destructor TBufferStream.Destroy;
begin
  if FHasWritten then WriteBuffer;
	FreeMem(FBuffer, FBufferSize);
  inherited Destroy;
end;

{ ************************* Contrle position ************************** }

procedure TBufferStream.SetSize(NewSize: LongInt);
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  FCurrentSize:= NewSize;
end;

function TBufferStream.Seek(Offset: LongInt; Origin: Word): LongInt;
begin
  if (Offset = 0) and (Origin = 2) then
  begin
    Result:= FCurrentSize;
    FHasReadSize:= True;
    Exit;
  end;
  FHasReadSize:= False;
  { Absolute Offset calculation }
  case Origin of
     0: begin end; { No Change : Offset:= Offset because already absolute }
     1: Offset:= GetCurrentPos + Offset;
     2: Offset:= FCurrentSize - Offset;
     else raise EStreamError.Create('L''origine de dplacement ' +
        IntToStr(Origin) + ' est indtermine');
  end;
  if Offset < 0 then Offset:= 0;
  { Do Seek }
  if FFilled then
  begin
     if (Offset >= FBufferPos) and (Offset <= (FBufferPos + FBufferSize)) then
     begin
        FVirtualPos:= Offset - FBufferPos;
        if FVirtualPos > FReadSize then FVirtualPos:= FReadSize;
     end else begin
	      if FHasWritten then WriteBuffer;
        FBufferPos:= Offset;
        FFilled:= False;
     end
  end else FBufferPos:= Offset;
  Result:= GetCurrentPos;
end;

function TBufferStream.GetBOF: Boolean;
begin
  Result:= GetCurrentPos = 0;
end;

function TBufferStream.GetEOF: Boolean;
begin
  Result:= GetCurrentPos = FCurrentSize;
end;

{ ********************************** Buffer ********************************** }

procedure TBufferStream.FlushBuffer;
begin
  WriteBuffer;
end;

procedure TBufferStream.InvalidateBuffer;
begin
  FFilled:= False;
  FCurrentSize:= GetSize;
end;

procedure TBufferStream.WriteBuffer;
var OldPos: Integer;
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  if FFilled then
  begin
     OldPos:= FVirtualPos;
     FVirtualPos:= 0;
     DirectWrite(FBuffer^, FReadSize);
     FVirtualPos:= OldPos;
     FCurrentSize:= GetSize; { Mise  jour de la taille si l'criture a chang qqchose }
  end;
  FHasWritten:= False;
end;

procedure TBufferStream.FillBuffer;
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  if FHasWritten then WriteBuffer; { Flush }
  { Pas besoin de remplir si c'est dj fait }
  if FFilled and (FVirtualPos = 0) then Exit;
  FReadSize:= DirectRead(FBuffer^, FBufferSize);
  FBufferPos:= GetCurrentPos;
  FVirtualPos:= 0;
  FFilled:= True;
end;

procedure TBufferStream.CheckSize;
begin
  { Si la VirtualPos est suprieure  la ReadSize, on met  jour }
	if (FVirtualPos > FReadSize) then FReadSize:= FVirtualPos;
end;

procedure TBufferStream.CheckEnd(Length: LongInt);
begin
  { Si on dpasse le buffer ou que celui-ci est vide ..., on le rerempli }
	if (((FVirtualPos + Length) >= FReadSize) and (FReadSize = FBufferSize)) or (not FFilled) then
		FillBuffer;
end;

procedure TBufferStream.CheckBuffer(Length: LongInt);
begin
  { Si on dpasse le buffer, alors on Reset }
	if ((FVirtualPos + Length) >= FBufferSize) or (not FFilled) then FillBuffer;
end;

function TBufferStream.GetCurrentPos: LongInt;
begin
  if FFilled then
     Result:= FBufferPos + FVirtualPos
  else Result:= FBufferPos;
end;

procedure TBufferStream.SetBufferSize(NewSize: Integer);
var OldPos: Integer;
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  if NewSize <= 0 then Exit;
  {$IFDEF WIN32} ReallocMem(FBuffer, NewSize);
  {$ELSE} ReallocMem(FBuffer, FBufferSize, NewSize); {$ENDIF}
  { Si on est dj rempli  mort, alors on complte ... sinon rien  faire }
  if FFilled and (NewSize > FBufferSize) and (FReadSize = FBufferSize) then
  begin
     FFilled:= False;
     OldPos:= FBufferPos;
     FBufferPos:= FBufferPos + FReadSize;
     Inc(FReadSize, DirectRead(Pointer(LongInt(FBuffer) + FBufferSize)^, NewSize - FBufferSize));
     FFilled:= True;
     FBufferPos:= OldPos;
  end;
  FBufferSize:= NewSize;
end;

{ ***************************** Lecture ****************************** }

function TBufferStream.Read(var Buffer; Count: LongInt): LongInt;
var
  L, NewPos: Integer;
  PBuff: Pointer;
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  Count:= IMin(FCurrentSize - GetCurrentPos, Count);{ Calcul du transfert rel }
  NewPos:= GetCurrentPos + Count;           { Nouvelle position }
  if Count >= FBufferSize then
     Result:= DirectRead(Buffer, Count)     { On crit directement dans la cible }
  else begin
     if not FFilled then FillBuffer;
     { Flush du buffer }
     PBuff:= Pointer(LongInt(FBuffer) + FVirtualPos);
     L:= IMin(FReadSize - FVirtualPos, Count); { RS - VP = Taille restante du buffer }
     Move(PBuff^, Buffer, L);               { Transfert des donnes }
     Inc(FVirtualPos, L);                   { On place VP l o on est }
     Result:= L;                            { Result = Count octets transfrs }
     Dec(Count, L);                         { Il y a moins de truc  rcuprer }
     if Count > 0 then
     begin
        FillBuffer;                            { Remplissage du buffer }
        PBuff:= Pointer(LongInt(@Buffer) + L); { Et on crira l o il faut (+ loin) }
        Move(FBuffer^, PBuff^, Count);         { Transfert de ce qui reste }
        Inc(Result, Count);
     end;
  end;
  Seek(NewPos, 0);                          { Mise  jour de la position }
end;

{ ********************************* criture ********************************* }

function TBufferStream.Write(const Buffer; Count: LongInt): LongInt;
var
  L, NewPos: Integer;
  PBuff: Pointer;
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  NewPos:= Position + Count;
  FCurrentSize:= IMax(Position + Count, Size);        { Mise  jour initiale }
  Result:= Count;
  if Count >= FBufferSize then
  begin
     if FHasWritten then WriteBuffer;
     DirectWrite(Buffer, Count);      { Note : le Seek final rglera la Pos }
     FCurrentSize:= GetSize; { Si la taille du truc a change, alors, on la met  jour }
  end else begin
     if not FFilled then FillBuffer;
     { L, on transferre tout ce qui peut/doit l'tre }
     PBuff:= Pointer(LongInt(FBuffer) + FVirtualPos);
     L:= IMin(FBufferSize - FVirtualPos, Count);  { BS - VP = Taille restante du buffer }
     Move(Buffer, PBuff^, L);                  { Transfert }
     Inc(FVirtualPos, L);                      { On place VP l o on est }
     Dec(Count, L);                            { Il y a moins de truc  crire }
     CheckSize;                                { On met  jour RS si ncessaire }
     if Count > 0 then
     begin
        FillBuffer;
        PBuff:= Pointer(Integer(@Buffer) + L);    { Ce qui reste  crire }
        Move(PBuff^, FBuffer^, Count);            { Transfert du reste }
        Inc(FVirtualPos, Count);                  { On place VP l o on est }
        CheckSize;
     end;
     FHasWritten:= True;                    { De toute faon, on a crit ! }
  end;
  Seek(NewPos, 0);                          { Mise  jour de la position }
end;

{ **************************** Fonctions I/O natives *************************** }

function TBufferStream.ReadChar: Char;
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  if not EOF then begin
	   CheckEnd(1);                           { Reremplie le Buffer si ncessaire }
     Result:= FBuffer^[FVirtualPos];
     Inc(FVirtualPos);
  end else Result:= Chr(0);
end;

procedure TBufferStream.ReadLn(var S : string);
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
	S:= '';
  while not (FBuffer^[FVirtualPos] in [Chr(0), Chr(13)]) do
  begin
     CheckEnd(1);
     if EOF then Exit;
     S:= S + FBuffer^[FVirtualPos];
     Inc(FVirtualPos);
  end;
  ReadChar; ReadChar;
end;

procedure TBufferStream.WriteChar(Char: Char);
begin
  if FHasReadSize then Seek(FCurrentSize, 0);
  CheckBuffer(1);
  FBuffer^[FVirtualPos]:= Char;
  Inc(FVirtualPos);
  FHasWritten:= True;
  if EOF then Inc(FCurrentSize);
  CheckSize;
end;

procedure TBufferStream.WriteLn(const Str: string);
begin
  Write(PChar(Str + #13#10)^, Length(Str) + 2);
end;

{ **************************************************************************** }
{ **************************** TFileBufferStream ***************************** }
{ **************************************************************************** }

constructor TFileBufferStream.Create(const FileName: string; Mode: Word);
begin
  if Mode = fmCreate then
    FHandle:= FileCreate(FileName)
  else
    FHandle:= FileOpen(FileName, Mode);
  if FHandle < 0 then
    raise EFileError.Create('Impossible d''ouvrir le fichier ' + FileName);
  inherited Create(32768);
end;

destructor TFileBufferStream.Destroy;
begin
  if FHasWritten then WriteBuffer;
  FileClose(FHandle);
  inherited;
end;

function TFileBufferStream.DirectRead(var Buffer; Count: LongInt): LongInt;
begin
  FileSeek(FHandle, Position, 0);
  Result:= FileRead(FHandle, Buffer, Count);
end;

procedure TFileBufferStream.DirectWrite(const Buffer; Count: LongInt);
begin
  FileSeek(FHandle, Position, 0);
  FileWrite(FHandle, Buffer, Count);
end;

function TFileBufferStream.GetSize: LongInt;
begin
  Result:= FileSeek(FHandle, 0, 2);
end;

end.

